@@ -1019,8 +1019,11 @@ filename_charset_cache_free (gpointer data)
  * subsequent character sets are used when trying to generate a displayable
  * representation of a filename, see g_filename_display_name().
  *
- * The character sets are determined by consulting the environment variables 
- * <envar>G_FILENAME_ENCODING</envar> and <envar>G_BROKEN_FILENAMES</envar>.
+ * On Unix, the character sets are determined by consulting the
+ * environment variables <envar>G_FILENAME_ENCODING</envar> and
+ * <envar>G_BROKEN_FILENAMES</envar>. On Windows, the character set
+ * used in the GLib API is always UTF-8 and said environment variables
+ * have no effect.
  *
  * <envar>G_FILENAME_ENCODING</envar> may be set to a comma-separated list 
  * of character set names. The special token "@locale" is taken to mean the 
@@ -1176,7 +1179,7 @@ _g_convert_thread_init (void)
  * @error:         location to store the error occuring, or %NULL to ignore
  *                 errors. Any of the errors in #GConvertError may occur.
  * 
- * Converts a string which is in the encoding used for filenames
+ * Converts a string which is in the encoding used by GLib for filenames
  * into a UTF-8 string.
  * 
  * Return value: The converted string, or %NULL on an error.
@@ -1694,6 +1697,29 @@ g_filename_from_uri (const gchar *uri,
   return result;
 }
 
+#ifdef G_OS_WIN32
+
+#undef g_filename_from_uri
+
+gchar *
+g_filename_from_uri (const gchar *uri,
+		     gchar      **hostname,
+		     GError     **error)
+{
+  gchar *utf8_filename;
+  gchar *retval = NULL;
+
+  utf8_filename = g_filename_from_uri_utf8 (uri, hostname, error);
+  if (utf8_filename)
+    {
+      retval = g_locale_from_utf8 (utf8_filename, -1, NULL, NULL, error);
+      g_free (utf8_filename);
+    }
+  return retval;
+}
+
+#endif
+
 /**
  * g_filename_to_uri:
  * @filename: an absolute filename specified in the encoding
@@ -1744,6 +1770,31 @@ g_filename_to_uri (const gchar *filename,
   return escaped_uri;
 }
 
+#ifdef G_OS_WIN32
+
+#undef g_filename_to_uri
+
+gchar *
+g_filename_to_uri (const gchar *filename,
+		   const gchar *hostname,
+		   GError     **error)
+{
+  gchar *utf8_filename;
+  gchar *retval = NULL;
+
+  utf8_filename = g_locale_to_utf8 (filename, -1, NULL, NULL, error);
+
+  if (utf8_filename)
+    {
+      retval = g_filename_to_uri_utf8 (utf8_filename, hostname, error);
+      g_free (utf8_filename);
+    }
+
+  return retval;
+}
+
+#endif
+
 /**
  * g_uri_list_extract_uris:
  * @uri_list: an URI list 
@@ -1855,6 +1906,40 @@ make_valid_utf8 (const gchar *name)
   return g_string_free (string, FALSE);
 }
 
+/**
+ * g_filename_display_basename:
+ * @filename: an absolute pathname in the GLib file name encoding
+ *
+ * Returns the display basename for the particular filename, guaranteed
+ * to be valid UTF-8. The display name might not be identical to the filename,
+ * for instance there might be problems converting it to UTF-8, and some files
+ * can be translated in the display
+ *
+ * You must pass the whole absolute pathname to this functions so that
+ * translation of well known locations can be done.
+ *
+ * This function is preferred over g_filename_display_name() if you know the
+ * whole path, as it allows translation.
+ *
+ * Return value: a newly allocated string containing
+ *   a rendition of the basename of the filename in valid UTF-8
+ *
+ * Since: 2.6
+ **/
+gchar *
+g_filename_display_basename (const gchar *filename)
+{
+  char *basename;
+  char *display_name;
+
+  g_return_val_if_fail (filename != NULL, NULL);
+  
+  basename = g_path_get_basename (filename);
+  display_name = g_filename_display_name (basename);
+  g_free (basename);
+  return display_name;
+}
+
 /**
  * g_filename_display_name:
  * @filename: a pathname hopefully in the GLib file name encoding
@@ -1867,6 +1952,10 @@ make_valid_utf8 (const gchar *name)
  * to be non-NULL even if the filename actually isn't in the GLib
  * file name encoding.
  *
+ * If you know the whole pathname of the file you should use
+ * g_filename_display_basename(), since that allows location-based
+ * translation of filenames.
+ *
  * Return value: a newly allocated string containing
  *   a rendition of the filename in valid UTF-8
  *
