@@ -19,9 +19,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include "goption.h"
+#include "config.h"
 
 #include "galias.h"
+
+#include "goption.h"
 #include "glib.h"
 #include "gi18n.h"
 
@@ -112,6 +114,19 @@ g_option_error_quark (void)
   return q;
 }
 
+/**
+ * g_option_context_new:
+ * @parameter_string: a string which is displayed in
+ *    the first line of <option>--help</option> output, after 
+ *    <literal><replaceable>programname</replaceable> [OPTION...]</literal>
+ *
+ * Creates a new option context. 
+ *
+ * Returns: a newly created #GOptionContext, which must be
+ *    freed with g_option_context_free() after use.
+ *
+ * Since: 2.6
+ */
 GOptionContext *
 g_option_context_new (const gchar *parameter_string)
 
@@ -122,12 +137,21 @@ g_option_context_new (const gchar *parameter_string)
 
   context->parameter_string = g_strdup (parameter_string);
   context->help_enabled = TRUE;
+  context->ignore_unknown = FALSE;
 
   return context;
 }
 
-void
-g_option_context_free (GOptionContext *context)
+/**
+ * g_option_context_free:
+ * @context: a #GOptionContext 
+ *
+ * Frees context and all the groups which have been 
+ * added to it.
+ *
+ * Since: 2.6
+ */
+void g_option_context_free (GOptionContext *context) 
 {
   g_return_if_fail (context != NULL);
 
@@ -145,9 +169,22 @@ g_option_context_free (GOptionContext *context)
   g_free (context);
 }
 
-void
-g_option_context_set_help_enabled (GOptionContext *context,
-				   gboolean        help_enabled)
+
+/**
+ * g_option_context_set_help_enabled:
+ * @context: a #GOptionContext
+ * @help_enabled: %TRUE to enable <option>--help</option>, %FALSE to disable it
+ *
+ * Enables or disables automatic generation of <option>--help</option> 
+ * output. By default, g_option_context_parse() recognizes
+ * <option>--help</option>, <option>-?</option>, <option>--help-all</option>
+ * and <option>--help-</option><replaceable>groupname</replaceable> and creates
+ * suitable output to stdout. 
+ *
+ * Since: 2.6
+ */
+void g_option_context_set_help_enabled (GOptionContext *context,
+                                        gboolean        help_enabled)
 
 {
   g_return_if_fail (context != NULL);
@@ -155,14 +192,41 @@ g_option_context_set_help_enabled (GOptionContext *context,
   context->help_enabled = help_enabled;
 }
 
-gboolean
-g_option_context_get_help_enabled (GOptionContext *context)
+/**
+ * g_option_context_get_help_enabled:
+ * @context: a #GOptionContext
+ * 
+ * Returns whether automatic <option>--help</option> generation
+ * is turned on for @context. See g_option_context_set_help_enabled().
+ * 
+ * Returns: %TRUE if automatic help generation is turned on.
+ *
+ * Since: 2.6
+ */
+gboolean 
+g_option_context_get_help_enabled (GOptionContext *context) 
 {
   g_return_val_if_fail (context != NULL, FALSE);
-
+  
   return context->help_enabled;
 }
 
+/**
+ * g_option_context_set_ignore_unknown_options:
+ * @context: a #GOptionContext
+ * @ignore_unknown: %TRUE to ignore unknown options, %FALSE to produce
+ *    an error when unknown options are met
+ * 
+ * Sets whether to ignore unknown options or not. If an argument is 
+ * ignored, it is left in the @argv array after parsing. By default, 
+ * g_option_context_parse() treats unknown options as error.
+ * 
+ * This setting does not affect non-option arguments (i.e. arguments 
+ * which don't start with a dash). But note that GOption cannot reliably
+ * determine whether a non-option belongs to a preceding unknown option.
+ *
+ * Since: 2.6
+ **/
 void
 g_option_context_set_ignore_unknown_options (GOptionContext *context,
 					     gboolean	     ignore_unknown)
@@ -172,6 +236,17 @@ g_option_context_set_ignore_unknown_options (GOptionContext *context,
   context->ignore_unknown = ignore_unknown;
 }
 
+/**
+ * g_option_context_get_ignore_unknown_options:
+ * @context: a #GOptionContext
+ * 
+ * Returns whether unknown options are ignored or not. See
+ * g_option_context_set_ignore_unknown_options().
+ * 
+ * Returns: %TRUE if unknown options are ignored.
+ * 
+ * Since: 2.6
+ **/
 gboolean
 g_option_context_get_ignore_unknown_options (GOptionContext *context)
 {
@@ -180,19 +255,56 @@ g_option_context_get_ignore_unknown_options (GOptionContext *context)
   return context->ignore_unknown;
 }
 
+/**
+ * g_option_context_add_group:
+ * @context: a #GOptionContext
+ * @group: the group to add
+ * 
+ * Adds a #GOptionGroup to the @context, so that parsing with @context
+ * will recognize the options in the group. Note that the group will
+ * be freed together with the context when g_option_context_free() is
+ * called, so you must not free the group yourself after adding it
+ * to a context.
+ *
+ * Since: 2.6
+ **/
 void
 g_option_context_add_group (GOptionContext *context,
 			    GOptionGroup   *group)
 {
+  GList *list;
+
   g_return_if_fail (context != NULL);
   g_return_if_fail (group != NULL);
   g_return_if_fail (group->name != NULL);
   g_return_if_fail (group->description != NULL);
   g_return_if_fail (group->help_description != NULL);
 
-  context->groups = g_list_prepend (context->groups, group);
+  for (list = context->groups; list; list = list->next)
+    {
+      GOptionGroup *g = (GOptionGroup *)list->data;
+
+      if ((group->name == NULL && g->name == NULL) ||
+	  (group->name && g->name && strcmp (group->name, g->name) == 0))
+	g_warning ("A group named \"%s\" is already part of this GOptionContext", 
+		   group->name);
+    }
+
+  context->groups = g_list_append (context->groups, group);
 }
 
+/**
+ * g_option_context_set_main_group:
+ * @context: a #GOptionContext
+ * @group: the group to set as main group
+ * 
+ * Sets a #GOptionGroup as main group of the @context. 
+ * This has the same effect as calling g_option_context_add_group(), 
+ * the only difference is that the options in the main group are 
+ * treated differently when generating <option>--help</option> output.
+ *
+ * Since: 2.6
+ **/
 void
 g_option_context_set_main_group (GOptionContext *context,
 				 GOptionGroup   *group)
@@ -203,6 +315,18 @@ g_option_context_set_main_group (GOptionContext *context,
   context->main_group = group;
 }
 
+/**
+ * g_option_context_get_main_group:
+ * @context: a #GOptionContext
+ * 
+ * Returns a pointer to the main group of @context.
+ * 
+ * Return value: the main group of @context, or %NULL if @context doesn't
+ *  have a main group. Note that group belongs to @context and should
+ *  not be modified or freed.
+ *
+ * Since: 2.6
+ **/
 GOptionGroup *
 g_option_context_get_main_group (GOptionContext *context)
 {
@@ -211,6 +335,19 @@ g_option_context_get_main_group (GOptionContext *context)
   return context->main_group;
 }
 
+/**
+ * g_option_context_add_main_entries:
+ * @context: a #GOptionContext
+ * @entries: a %NULL-terminated array of #GOptionEntry<!-- -->s
+ * @translation_domain: a translation domain to use for translating
+ *    the <option>--help</option> output for the options in @entries
+ *    with gettext(), or %NULL
+ * 
+ * A convenience function which creates a main group if it doesn't 
+ * exist, adds the @entries to it and sets the translation domain.
+ * 
+ * Since: 2.6
+ **/
 void
 g_option_context_add_main_entries (GOptionContext      *context,
 				   const GOptionEntry  *entries,
@@ -234,7 +371,7 @@ print_entry (GOptionGroup       *group,
 
   if (entry->flags & G_OPTION_FLAG_HIDDEN)
     return;
-	  
+
   str = g_string_new (NULL);
   
   if (entry->short_name)
@@ -258,11 +395,52 @@ print_help (GOptionContext *context,
   GList *list;
   gint max_length, len;
   gint i;
+  GHashTable *shadow_map;
+  gboolean seen[256];
   
   g_print ("%s\n  %s %s %s\n\n", 
 	   _("Usage:"), g_get_prgname(), _("[OPTION...]"),
 	   context->parameter_string ? context->parameter_string : "");
 
+  memset (seen, 0, sizeof (gboolean) * 256);
+  shadow_map = g_hash_table_new (g_str_hash, g_str_equal);
+
+  if (context->main_group)
+    {
+      for (i = 0; i < context->main_group->n_entries; i++)
+	{
+	  g_hash_table_insert (shadow_map, 
+			       (gpointer)context->main_group->entries[i].long_name, 
+			       context->main_group->entries + i);
+	  
+	  if (seen[(guchar)context->main_group->entries[i].short_name])
+	    context->main_group->entries[i].short_name = 0;
+	  else
+	    seen[(guchar)context->main_group->entries[i].short_name] = TRUE;
+	}
+    }
+
+  list = context->groups;
+  while (list != NULL)
+    {
+      GOptionGroup *group = list->data;
+      for (i = 0; i < group->n_entries; i++)
+	{
+	  if (g_hash_table_lookup (shadow_map, group->entries[i].long_name))
+	    group->entries[i].long_name = g_strdup_printf ("%s-%s", group->name, group->entries[i].long_name);
+	  else  
+	    g_hash_table_insert (shadow_map, (gpointer)group->entries[i].long_name, group->entries + i);
+
+	  if (seen[(guchar)group->entries[i].short_name])
+	    group->entries[i].short_name = 0;
+	  else
+	    seen[(guchar)group->entries[i].short_name] = TRUE;
+	}
+      list = list->next;
+    }
+
+  g_hash_table_destroy (shadow_map);
+
   list = context->groups;
 
   max_length = g_utf8_strlen ("--help, -?", -1);
@@ -284,6 +462,9 @@ print_help (GOptionContext *context,
       /* Then we go through the entries */
       for (i = 0; i < group->n_entries; i++)
 	{
+	  if (group->entries[i].flags & G_OPTION_FLAG_HIDDEN)
+	    continue;
+
 	  len = g_utf8_strlen (group->entries[i].long_name, -1);
 
 	  if (group->entries[i].short_name)
@@ -301,26 +482,29 @@ print_help (GOptionContext *context,
 
   /* Add a bit of padding */
   max_length += 4;
-  
-  list = context->groups;
 
-  g_print ("%s\n  --%-*s %s\n", 
-	   _("Help Options:"), max_length, "help", _("Show help options"));
-
-  /* We only want --help-all when there are groups */
-  if (list)
-    g_print ("  --%-*s %s\n", max_length, "help-all", _("Show all help options"));
-
-  while (list)
+  if (!group)
     {
-      GOptionGroup *group = list->data;
-
-      g_print ("  --help-%-*s %s\n", max_length - 5, group->name, TRANSLATE (group, group->help_description));
+      list = context->groups;
       
-      list = list->next;
-    }
+      g_print ("%s\n  --%-*s %s\n", 
+	       _("Help Options:"), max_length, "help", _("Show help options"));
+      
+      /* We only want --help-all when there are groups */
+      if (list)
+	g_print ("  --%-*s %s\n", max_length, "help-all", _("Show all help options"));
 
-  g_print ("\n");
+      while (list)
+	{
+	  GOptionGroup *group = list->data;
+	  
+	  g_print ("  --help-%-*s %s\n", max_length - 5, group->name, TRANSLATE (group, group->help_description));
+	  
+	  list = list->next;
+	}
+
+      g_print ("\n");
+    }
 
   if (group)
     {
@@ -377,7 +561,6 @@ print_help (GOptionContext *context,
 
       g_print ("\n");
     }
-
   
   exit (0);
 }
@@ -474,7 +657,7 @@ parse_arg (GOptionContext *context,
 	change = get_change (context, G_OPTION_ARG_NONE,
 			     entry->arg_data);
 
-	*(gboolean *)entry->arg_data = TRUE;
+	*(gboolean *)entry->arg_data = !(entry->flags & G_OPTION_FLAG_REVERSE);
  	break;
       }	     
     case G_OPTION_ARG_STRING:
@@ -532,8 +715,14 @@ parse_arg (GOptionContext *context,
       {
 	gchar *data;
 
+#ifdef G_OS_WIN32
+	data = g_locale_to_utf8 (value, -1, NULL, NULL, error);
+	
+	if (!data)
+	  return FALSE;
+#else
 	data = g_strdup (value);
-
+#endif
 	change = get_change (context, G_OPTION_ARG_FILENAME,
 			     entry->arg_data);
 	g_free (change->allocated.str);
@@ -549,8 +738,14 @@ parse_arg (GOptionContext *context,
       {
 	gchar *data;
 	
+#ifdef G_OS_WIN32
+	data = g_locale_to_utf8 (value, -1, NULL, NULL, error);
+	
+	if (!data)
+	  return FALSE;
+#else
 	data = g_strdup (value);
-
+#endif
 	change = get_change (context, G_OPTION_ARG_STRING_ARRAY,
 			     entry->arg_data);
 
@@ -742,6 +937,40 @@ parse_long_option (GOptionContext *context,
   return TRUE;
 }
 
+static gboolean
+parse_remaining_arg (GOptionContext *context,
+		     GOptionGroup   *group,
+		     gint           *index,
+		     gint           *argc,
+		     gchar        ***argv,
+		     GError        **error,
+		     gboolean       *parsed)
+{
+  gint j;
+
+  for (j = 0; j < group->n_entries; j++)
+    {
+      if (*index >= *argc)
+	return TRUE;
+
+      if (group->entries[j].long_name[0])
+	continue;
+
+      g_return_val_if_fail (group->entries[j].arg == G_OPTION_ARG_STRING_ARRAY ||
+			    group->entries[j].arg == G_OPTION_ARG_FILENAME_ARRAY, FALSE);
+      
+      add_pending_null (context, &((*argv)[*index]), NULL);
+      
+      if (!parse_arg (context, group, &group->entries[j], (*argv)[*index], "", error))
+	return FALSE;
+      
+      *parsed = TRUE;
+      return TRUE;
+    }
+
+  return TRUE;
+}
+
 static void
 free_changes_list (GOptionContext *context,
 		   gboolean        revert)
@@ -814,6 +1043,34 @@ free_pending_nulls (GOptionContext *context,
   context->pending_nulls = NULL;
 }
 
+/**
+ * g_option_context_parse:
+ * @context: a #GOptionContext
+ * @argc: a pointer to the number of command line arguments.
+ * @argv: a pointer to the array of command line arguments.
+ * @error: a return location for errors 
+ * 
+ * Parses the command line arguments, recognizing options
+ * which have been added to @context. A side-effect of 
+ * calling this function is that g_set_prgname() will be
+ * called.
+ *
+ * If the parsing is successful, any parsed arguments are
+ * removed from the array and @argc and @argv are updated 
+ * accordingly. In case of an error, @argc and @argv are
+ * left unmodified.
+ *
+ * If automatic <option>--help</option> support is enabled
+ * (see g_option_context_set_help_enabled()), and the 
+ * @argv array contains one of the recognized help options,
+ * this function will produce help output to stdout and
+ * call <literal>exit (0)</literal>.
+ * 
+ * Return value: %TRUE if the parsing was successful, 
+ *               %FALSE if an error occurred
+ *
+ * Since: 2.6
+ **/
 gboolean
 g_option_context_parse (GOptionContext   *context,
 			gint             *argc,
@@ -862,12 +1119,14 @@ g_option_context_parse (GOptionContext   *context,
 
   if (argc && argv)
     {
+      gboolean stop_parsing = FALSE;
+
       for (i = 1; i < *argc; i++)
 	{
-	  gchar *arg;
+	  gchar *arg, *dash;
 	  gboolean parsed = FALSE;
 
-	  if ((*argv)[i][0] == '-')
+	  if ((*argv)[i][0] == '-' && !stop_parsing)
 	    {
 	      if ((*argv)[i][1] == '-')
 		{
@@ -879,7 +1138,8 @@ g_option_context_parse (GOptionContext   *context,
 		  if (*arg == 0)
 		    {
 		      add_pending_null (context, &((*argv)[i]), NULL);
-		      break;
+		      stop_parsing = TRUE;
+		      continue;
 		    }
 
 		  /* Handle help options */
@@ -930,12 +1190,41 @@ g_option_context_parse (GOptionContext   *context,
 		      
 		      list = list->next;
 		    }
+		  
+		  if (parsed)
+		    continue;
 
+		  /* Now look for --<group>-<option> */
+		  dash = strchr (arg, '-');
+		  if (dash)
+		    {
+		      /* Try the groups */
+		      list = context->groups;
+		      while (list)
+			{
+			  GOptionGroup *group = list->data;
+			  
+			  if (strncmp (group->name, arg, dash - arg) == 0)
+			    {
+			      if (!parse_long_option (context, group, &i, dash + 1,
+						      argc, argv, error, &parsed))
+				goto fail;
+			      
+			      if (parsed)
+				break;
+			    }
+			  
+			  list = list->next;
+			}
+		    }
+		  
 		  if (context->ignore_unknown)
 		    continue;
 		}
 	      else
 		{
+		  /* short option */
+
 		  gint new_i, j;
 		  gboolean *nulled_out = NULL;
 		  
@@ -1025,6 +1314,15 @@ g_option_context_parse (GOptionContext   *context,
 		  goto fail;
 		}
 	    }
+	  else
+	    {
+	      /* Collect remaining args */
+	      if (context->main_group &&
+		  !parse_remaining_arg (context, context->main_group, &i,
+					argc, argv, error, &parsed))
+		goto fail;
+	      
+	    }
 	}
 
       /* Call post-parse hooks */
@@ -1062,7 +1360,10 @@ g_option_context_parse (GOptionContext   *context,
 	    {
 	      k -= i;
 	      for (j = i + k; j < *argc; j++)
-		(*argv)[j-k] = (*argv)[j];
+		{
+		  (*argv)[j-k] = (*argv)[j];
+		  (*argv)[j] = NULL;
+		}
 	      *argc -= k;
 	    }
 	}      
@@ -1095,7 +1396,27 @@ g_option_context_parse (GOptionContext   *context,
   return FALSE;
 }
 				   
-     
+/**
+ * g_option_group_new:
+ * @name: the name for the option group, this is used to provide
+ *   help for the options in this group with <option>--help-</option>@name
+ * @description: a description for this group to be shown in 
+ *   <option>--help</option>. This string is translated using the translation
+ *   domain or translation function of the group
+ * @help_description: a description for the <option>--help-</option>@name option.
+ *   This string is translated using the translation domain or translation function
+ *   of the group
+ * @user_data: user data that will be passed to the pre- and post-parse hooks,
+ *   the error hook and to callbacks of %G_OPTION_ARG_CALLBACK options, or %NULL
+ * @destroy: a function that will be called to free @user_data, or %NULL
+ * 
+ * Creates a new #GOptionGroup.
+ * 
+ * Return value: a newly created option group. It should be added 
+ *   to a #GOptionContext or freed with g_option_group_free().
+ *
+ * Since: 2.6
+ **/
 GOptionGroup *
 g_option_group_new (const gchar    *name,
 		    const gchar    *description,
@@ -1116,6 +1437,16 @@ g_option_group_new (const gchar    *name,
   return group;
 }
 
+
+/**
+ * g_option_group_free:
+ * @group: a #GOptionGroup
+ * 
+ * Frees a #GOptionGroup. Note that you must <emphasis>not</emphasis>
+ * free groups which have been added to a #GOptionContext.
+ *
+ * Since: 2.6
+ **/
 void
 g_option_group_free (GOptionGroup *group)
 {
@@ -1137,6 +1468,15 @@ g_option_group_free (GOptionGroup *group)
 }
 
 
+/**
+ * g_option_group_add_entries:
+ * @group: a #GOptionGroup
+ * @entries: a %NULL-terminated array of #GOptionEntry<!-- -->s
+ * 
+ * Adds the options specified in @entries to @group.
+ *
+ * Since: 2.6
+ **/
 void
 g_option_group_add_entries (GOptionGroup       *group,
 			    const GOptionEntry *entries)
@@ -1154,6 +1494,22 @@ g_option_group_add_entries (GOptionGroup       *group,
   group->n_entries += n_entries;
 }
 
+/**
+ * g_option_group_set_parse_hooks:
+ * @group: a #GOptionGroup
+ * @pre_parse_func: a function to call before parsing, or %NULL
+ * @post_parse_func: a function to call after parsing, or %NULL
+ * 
+ * Associates two functions with @group which will be called 
+ * from g_option_context_parse() before the first option is parsed
+ * and after the last option has been parsed, respectively.
+ *
+ * Note that the user data to be passed to @pre_parse_func and
+ * @post_parse_func can be specified when constructing the group
+ * with g_option_group_new().
+ *
+ * Since: 2.6
+ **/
 void
 g_option_group_set_parse_hooks (GOptionGroup     *group,
 				GOptionParseFunc  pre_parse_func,
@@ -1165,6 +1521,20 @@ g_option_group_set_parse_hooks (GOptionGroup     *group,
   group->post_parse_func = post_parse_func;  
 }
 
+/**
+ * g_option_group_set_error_hook:
+ * @group: a #GOptionGroup
+ * @error_func: a function to call when an error occurs
+ * 
+ * Associates a function with @group which will be called 
+ * from g_option_context_parse() when an error occurs.
+ *
+ * Note that the user data to be passed to @pre_parse_func and
+ * @post_parse_func can be specified when constructing the group
+ * with g_option_group_new().
+ *
+ * Since: 2.6
+ **/
 void
 g_option_group_set_error_hook (GOptionGroup     *group,
 			       GOptionErrorFunc	 error_func)
@@ -1175,11 +1545,28 @@ g_option_group_set_error_hook (GOptionGroup     *group,
 }
 
 
+/**
+ * g_option_group_set_translate_func:
+ * @group: a #GOptionGroup
+ * @func: the #GTranslateFunc, or %NULL 
+ * @data: user data to pass to @func, or %NULL
+ * @destroy_notify: a function which gets called to free @data, or %NULL
+ * 
+ * Sets the function which is used to translate user-visible
+ * strings, for <option>--help</option> output. Different
+ * groups can use different #GTranslateFunc<!-- -->s. If @func
+ * is %NULL, strings are not translated.
+ *
+ * If you are using gettext(), you only need to set the translation
+ * domain, see g_option_group_set_translation_domain().
+ *
+ * Since: 2.6
+ **/
 void
 g_option_group_set_translate_func (GOptionGroup   *group,
 				   GTranslateFunc  func,
 				   gpointer        data,
-				   GDestroyNotify  notify)
+				   GDestroyNotify  destroy_notify)
 {
   g_return_if_fail (group != NULL);
   
@@ -1188,7 +1575,7 @@ g_option_group_set_translate_func (GOptionGroup   *group,
       
   group->translate_func = func;
   group->translate_data = data;
-  group->translate_notify = notify;
+  group->translate_notify = destroy_notify;
 }
 
 static gchar *
@@ -1198,6 +1585,16 @@ dgettext_swapped (const gchar *msgid,
   return dgettext (domainname, msgid);
 }
 
+/**
+ * g_option_group_set_translation_domain:
+ * @group: a #GOptionGroup
+ * @domain: the domain to use
+ * 
+ * A convenience function to use gettext() for translating
+ * user-visible strings. 
+ * 
+ * Since: 2.6
+ **/
 void
 g_option_group_set_translation_domain (GOptionGroup *group,
 				       const gchar  *domain)
