@@ -26,11 +26,9 @@
 #include <wchar.h>
 #endif
 
+#include "galias.h"
 #include "glib.h"
-
-extern gunichar *_g_utf8_normalize_wc (const gchar    *str,
-				       gssize          max_len,
-				       GNormalizeMode  mode);
+#include "gunicodeprivate.h"
 
 /**
  * g_utf8_collate:
@@ -41,7 +39,7 @@ extern gunichar *_g_utf8_normalize_wc (const gchar    *str,
  * correct rules for the current locale. When sorting a large
  * number of strings, it will be significantly faster to
  * obtain collation keys with g_utf8_collate_key() and 
- * compare the keys with <function>strcmp()</function> when 
+ * compare the keys with strcmp() when 
  * sorting instead of sorting the original strings.
  * 
  * Return value: -1 if @str1 compares before @str2, 0 if they
@@ -87,8 +85,8 @@ g_utf8_collate (const gchar *str1,
     }
   else
     {
-      gchar *str1_locale = g_convert (str1_norm, -1, "UTF-8", charset, NULL, NULL, NULL);
-      gchar *str2_locale = g_convert (str2_norm, -1, "UTF-8", charset, NULL, NULL, NULL);
+      gchar *str1_locale = g_convert (str1_norm, -1, charset, "UTF-8", NULL, NULL, NULL);
+      gchar *str2_locale = g_convert (str2_norm, -1, charset, "UTF-8", NULL, NULL, NULL);
 
       if (str1_locale && str2_locale)
 	result =  strcoll (str1_locale, str2_locale);
@@ -166,9 +164,9 @@ utf8_encode (char *buf, wchar_t val)
  * @len: length of @str, in bytes, or -1 if @str is nul-terminated.
  *
  * Converts a string into a collation key that can be compared
- * with other collation keys using <function>strcmp()</function>. 
+ * with other collation keys using strcmp(). 
  * The results of comparing the collation keys of two strings 
- * with <function>strcmp()</function> will always be the same as 
+ * with strcmp() will always be the same as 
  * comparing the two original keys with g_utf8_collate().
  * 
  * Return value: a newly allocated string. This string should
@@ -229,11 +227,19 @@ g_utf8_collate_key (const gchar *str,
     }
   else
     {
-      gchar *str_locale = g_convert (str_norm, -1, "UTF-8", charset, NULL, NULL, NULL);
+      gchar *str_locale = g_convert (str_norm, -1, charset, "UTF-8", NULL, NULL, NULL);
 
       if (str_locale)
 	{
 	  xfrm_len = strxfrm (NULL, str_locale, 0);
+	  if (xfrm_len < 0 || xfrm_len >= G_MAXINT - 2)
+	    {
+	      g_free (str_locale);
+	      str_locale = NULL;
+	    }
+	}
+      if (str_locale)
+	{
 	  result = g_malloc (xfrm_len + 2);
 	  result[0] = 'A';
 	  strxfrm (result + 1, str_locale, xfrm_len + 1);
